// CardgameServer.cpp : Defines the entry point for the console application.
//

#include "CardgameServer.hpp"
#include "ConfigManager.hpp"
#include "packets/NetworkCommandIdentifiers.hpp"

#include <boost/bind.hpp>

#include <iostream>
#include <typeinfo>
#include <set>

#define RECEIVE_BUFFER_SIZE 128

Player::Player(session* p_session) :
	ritualPoints(0),
	mana(0),
	networkSession(p_session) {
}


GameSession::GameSession(CardManager* p_cardManager) {
	currentPlayer = 0;

	players[0] = nullptr;
	players[1] = nullptr;

	m_currentState = EGameState::WAIT_FOR_PLAYERS;
	m_cardManager = p_cardManager;
};

void GameSession::playerDisconnected(Player* p_player)
{
	// End the game unless game is already ending
	if( m_currentState != EGameState::GAME_END )
	{
		currentPlayer = 1 - p_player->serverSlot; //hax
		gameEnd();
	}
}

void GameSession::handlepacket(Player* p_player, nlohmann::json json_packet)
{
	std::string action = json_packet["id"].get<std::string>();
	std::cout << "Action: " << action << std::endl;
	if (action == "cl_keepalive") {
		sendKeepalive(p_player->serverSlot);
	}

	if( m_currentState == EGameState::PLAYER_ACTION && p_player == players[currentPlayer] )
	{
		playerActions(action, json_packet);
	}
}

bool GameSession::joinGame(Player* player) {

	// Try to add the player to the game. This sould always succeed.
	for (int i = 0; i < 2; i++) {
		if (players[i] == nullptr) {
			player->serverSlot = i;
			players[i] = player;
			//sendCardDump(i);
			notifyJoined(i);
			if (isGameFull() == true)
				startGame();
			return true;
		}
	}

	return false;
}

void GameSession::startGame() {
	m_currentState = EGameState::INITIALIZE_GAME;

	// Initalize players

	// Create random decks
	for (int i = 0; i < 30; i++)
		players[0]->deck.push_back(m_cardManager->getRandom());

	for (int i = 0; i < 30; i++)
		players[1]->deck.push_back(m_cardManager->getRandom());

	// Initialize field data structures: id -1 describes an empty slot

	Card c;
	c.id = -1;

	for (int i = 0; i < 7; i++)
		players[0]->field.push_back(c);

	for (int i = 0; i < 7; i++)
		players[1]->field.push_back(c);

	//    currentPlayer = math.random??	// Randomize starting player

	for (int cardNum = 0; cardNum < 4; cardNum++) {
		for (int i = 0; i < 2; i++) {

			/*
			Card drawn = players[i]->deck.back();		// Limited hand
			players[i]->deck.pop_back();
			players[i]->hand.push_back(drawn);
			/*/
			Card drawn = m_cardManager->getRandom();		// Infinite hand
			players[i]->hand.push_back(drawn);
			//*/
			notifyCardDraw(i, drawn);
		}
	}

	// Go to player_turn_start state
	playerStartTurn();
}

void GameSession::playerStartTurn() {
	// Increment win condition
	Player* player = players[currentPlayer];

	bool statsChanged = false;
	for (int i = 0; i < 8; i = i + 2) {	// Check back row
		// Get the card from the field
		Card c = player->field[i];

		// If card id valid
		if (c.id > -1) {
			players[currentPlayer]->ritualPoints += c.rp;
			statsChanged = true;
		}
	}

	players[currentPlayer]->mana = ConfigManager::manaMax;

	//if (statsChanged) {
		notifyUpdatePlayer(currentPlayer, players[currentPlayer]);
	//}

	// Check win condition
	if (players[currentPlayer]->ritualPoints >= ConfigManager::ritualMax) {
		//m_currentState = EGameState::GAME_END;
		gameEnd();

		return;
	}

	// Draw a card for the player


	/*
	Card drawn = player->deck.back();			// Limited hand
	player->deck.pop_back();
	player->hand.push_back(drawn);
	/*/
	Card drawn = m_cardManager->getRandom();		// Infinite hand
	player->hand.push_back(drawn);
	//*/

	notifyCardDraw(currentPlayer, drawn);

	// Experimental: in case of overdraw 10 cards? destroy new card
	if( player->hand.size() > ConfigManager::handSize ) {
		notifyDiscardCard(currentPlayer, 7 + player->hand.size()-1);
		player->hand.erase(player->hand.begin() + player->hand.size() - 1);
	}


	// Reset card action points
	for (int i = 0; i < 7; i++) {
		player->field[i].actionPoints = 1;
	}

	// Notify the player that the turn has started, allowing it to perform actions
	m_currentState = EGameState::PLAYER_ACTION;
	notifyTurnStart(currentPlayer);
}

void GameSession::playerActions(const std::string& action, nlohmann::json json_object) {

	if( action == PacketIdentifier::CLIENT_PLAY_CARD )
	{
		// TODO Check if move is valid

		int clientMoveId = json_object["moveId"];
		int sourceSlotId = json_object["sourceSlot"];
		int targetSlotId = json_object["targetSlot"];
		bool targetPlayer = json_object["targetPlayer"];

		// To simplify, first check for spells
		if (sourceSlotId >= 7) {  // In hand
			try {
				Card card = players[currentPlayer]->hand.at(sourceSlotId-7);
				if (card.type == (int)(CardType::Spell)) {
					if (targetSlotId == -1 && targetPlayer == true) {
						// Check mana cost
						if (players[currentPlayer]->mana >= card.manaCost) {
							players[currentPlayer]->mana -= card.manaCost;
							if (card.manaCost != 0) {
								notifyUpdatePlayer(currentPlayer, players[currentPlayer]);
							}
						}
						else {
							notifyErrorMessage(currentPlayer, "invalid_move", clientMoveId);
							return;
						}

						if (card.script == "draw4") {
							notifyPlayCard(currentPlayer, currentPlayer, sourceSlotId, targetSlotId, clientMoveId);
							for (int i = 0; i < 4; i++) {
								Card drawn = m_cardManager->getRandom();
								players[currentPlayer]->hand.push_back(drawn);
								notifyCardDraw(currentPlayer, drawn);
								
								if (players[currentPlayer]->hand.size() > ConfigManager::handSize) {
									notifyDiscardCard(currentPlayer, 7 + players[currentPlayer]->hand.size() - 1);
									players[currentPlayer]->hand.erase(players[currentPlayer]->hand.begin() + players[currentPlayer]->hand.size() - 1);
								}
							}
							return;
						}
					}
					notifyErrorMessage(currentPlayer, "invalid_move", clientMoveId);
					return;
				}
			}
			catch (std::exception& e) {
				std::cerr << "No such card id found from hand" << std::endl;
				notifyErrorMessage(currentPlayer, "invalid_move", clientMoveId);
				return;
			}
		}

		
		// is move hand -> field
		if( 7 <= sourceSlotId )
		{
			try
			{
				Card cardAtField = players[currentPlayer]->field.at(targetSlotId);

				if( cardAtField.id == -1 )
				{
					// Get and remove the ard from hand
					Card cardFromHand = players[currentPlayer]->hand.at(sourceSlotId-7);

					// Check mana cost
					if (players[currentPlayer]->mana >= cardFromHand.manaCost) {
						players[currentPlayer]->mana -= cardFromHand.manaCost;
						if (cardFromHand.manaCost != 0) {
							notifyUpdatePlayer(currentPlayer, players[currentPlayer]);
						}
					}
					else {
						notifyErrorMessage(currentPlayer, "invalid_move", clientMoveId);
						return;
					}

					players[currentPlayer]->hand.erase( players[currentPlayer]->hand.begin() + (sourceSlotId-7) );

					// Actually place the card to the field
					players[currentPlayer]->field[targetSlotId] = cardFromHand;

					// Send play card to both parties
					notifyPlayCard(currentPlayer, currentPlayer, sourceSlotId, targetSlotId, clientMoveId);
					
					if (cardFromHand.script == "draw3") {
						for (int i = 0; i < 3; i++) {
							Card drawn = m_cardManager->getRandom();
							players[currentPlayer]->hand.push_back(drawn);
							notifyCardDraw(currentPlayer, drawn);

							if (players[currentPlayer]->hand.size() > ConfigManager::handSize) {
								notifyDiscardCard(currentPlayer, 7 + players[currentPlayer]->hand.size() - 1);
								players[currentPlayer]->hand.erase(players[currentPlayer]->hand.begin() + players[currentPlayer]->hand.size() - 1);
							}
						}
					}

					if (cardFromHand.script == "charge") {
						players[currentPlayer]->field[targetSlotId].actionPoints = 1;
					}
					
					return;
				}

			}
			catch(std::exception& e)
			{
				std::cerr << "Invalid card id or something." << std::endl;
			}
		}

		// is move field -> field
		if( 0 <= sourceSlotId && sourceSlotId <= 6)
		if( targetPlayer == true )
		{
			Card sourceCard = players[currentPlayer]->field.at(sourceSlotId);
			Card targetCard = players[currentPlayer]->field.at(targetSlotId);

			// is source and target  valid
			if( sourceCard.id > -1 && targetCard.id == -1 )
			{
				// move card ( switch cars )
				players[currentPlayer]->field[sourceSlotId] = targetCard;
				players[currentPlayer]->field[targetSlotId] = sourceCard;

				// Notify movenotifyPlayCard
				// TODO
				notifyPlayCard(currentPlayer, currentPlayer, sourceSlotId, targetSlotId, clientMoveId);
				return;
			}
		}

		// is move attack field -> enemy card
		if( 0 <= sourceSlotId && sourceSlotId <= 6)
		if( targetPlayer == false ) {
			Card sourceCard = players[currentPlayer]->field.at(sourceSlotId);
			Card targetCard = players[1-currentPlayer]->field.at(targetSlotId);

			if (sourceCard.actionPoints > 0) {
				// Action points decreast at the end
			}
			else {
				// Nice control flow
				// Reason: no action points left
				notifyErrorMessage(currentPlayer, "invalid_move", clientMoveId);
				return;
			}

			// is source and target  valid
			if( sourceCard.id > -1 && targetCard.id > -1 )
			{
				// Check if the slot is protected
				if (targetSlotId % 2 == 0) { // Only for backline
					int slotA = targetSlotId - 1;
					int slotB = targetSlotId + 1;

					// Combined bounds check and slot occupation check
					// The slot is empty if OoB or the card id is -1. Negate that to get protection.
					bool protA = !(slotA == -1 || players[1-currentPlayer]->field.at(slotA).id == -1);
					bool protB = !(slotB ==  7 || players[1-currentPlayer]->field.at(slotB).id == -1);

					if (protA || protB) {
						// Nice control flow
						notifyErrorMessage(currentPlayer, "invalid_move", clientMoveId);
						return;
					}
				}

				// do attack
				int sourceHP = players[currentPlayer]->field[sourceSlotId].hp;
				int targetHP = players[1-currentPlayer]->field[targetSlotId].hp;

				sourceHP -= targetCard.attack;
				targetHP -= sourceCard.attack;

				players[currentPlayer]->field[sourceSlotId].hp = sourceHP;
				players[1-currentPlayer]->field[targetSlotId].hp = targetHP;

				notifyPlayCard(currentPlayer, 1-currentPlayer, sourceSlotId, targetSlotId, clientMoveId);

				notifyUpdateCard(currentPlayer, sourceSlotId, players[currentPlayer]->field[sourceSlotId]);
				notifyUpdateCard(1-currentPlayer, targetSlotId, players[1-currentPlayer]->field[targetSlotId]);

				// did anything die
				if( sourceHP <= 0 ) {
					notifyDiscardCard(currentPlayer, sourceSlotId);
					players[currentPlayer]->field[sourceSlotId].id = -1;
				}
				if( targetHP <= 0 ) {
					notifyDiscardCard(1-currentPlayer, targetSlotId);
					players[1-currentPlayer]->field[targetSlotId].id = -1;

					// Increment RPs for the attacker as the target died
					int rp = players[1 - currentPlayer]->field[targetSlotId].rp;
					players[currentPlayer]->ritualPoints += rp;
					if (rp != 0) { // If there is a change
						notifyUpdatePlayer(currentPlayer, players[currentPlayer]);
					}
				}

				players[currentPlayer]->field[sourceSlotId].actionPoints--;
				std::cout << "AP after, is it 0?: " << players[currentPlayer]->field[sourceSlotId].actionPoints << std::endl;

				return;
			}
		}

		// Action was invalid
		notifyErrorMessage(currentPlayer, "invalid_move", clientMoveId);
	}
	else if( action == PacketIdentifier::CLIENT_TURN_END )
	{
		playerEndTurn();
	}
	else if( action == PacketIdentifier::CLIENT_CONCEDE )
	{
		// TODO Concede
		players[1-currentPlayer]->ritualPoints = 1000000;	// TODO do better? // Set opponents ritual points so that they win
		playerEndTurn();
	}
}

void GameSession::playerEndTurn() {
	notifyTurnEnd(currentPlayer);
	changeTurn();
}

void GameSession::changeTurn() {
	if (currentPlayer == 0)
		currentPlayer = 1;
	else
		currentPlayer = 0;
	playerStartTurn();
}

void GameSession::gameEnd() {
	// TODO paremmin
	m_currentState = EGameState::GAME_END;

	notifyWinner(currentPlayer);

	if( players[0] != nullptr )
	{
		std::cout << "Disconnect player 0" << std::endl;
		players[0]->networkSession->disconnect();
		players[0] = nullptr;
	}

	if( players[1] != nullptr )
	{
		std::cout << "Disconnect player 1" << std::endl;
		players[1]->networkSession->disconnect();
		players[1] = nullptr;
	}

	m_currentState = EGameState::WAIT_FOR_PLAYERS;
}

void GameSession::cleanup() {
	//	players
}

void GameSession::notifyUpdatePlayer(int playerId, Player* player) {
	for (int i = 0; i < 2; i++) {
		nlohmann::json statesPayload;
		statesPayload["ritual"] = player->ritualPoints;
		statesPayload["mana"] = player->mana;

		nlohmann::json packet;
		packet["id"] = PacketIdentifier::SERVER_UPDATE_PLAYER;
		packet["states"] = statesPayload;
		packet["player"] = (playerId == i);

		sendNotify(i, packet);
	}
}

void GameSession::notifyJoined(int playerId) {
	for (int i = 0; i < 2; i++) {
		nlohmann::json packet;
		packet["id"] = PacketIdentifier::SERVER_CLIENT_JOINED;
		packet["player"] = (playerId == i);

		sendNotify(i, packet);
	}
}

void GameSession::notifyTurnStart(int playerId) {
	for (int i = 0; i < 2; i++) {
		nlohmann::json packet;
		packet["id"] = PacketIdentifier::SERVER_TURN_START;
		packet["player"] = (playerId == i);

		sendNotify(i, packet);
	}
}

void GameSession::notifyTurnEnd(int playerId) {
	for (int i = 0; i < 2; i++) {
		nlohmann::json packet;
		packet["id"] = PacketIdentifier::SERVER_TURN_END;
		packet["player"] = (playerId == i);

		sendNotify(i, packet);
	}
}

void GameSession::notifyCardDraw(int playerId, const Card& card) {
	for (int i = 0; i < 2; i++) {
		nlohmann::json packet;
		packet["id"] = PacketIdentifier::SERVER_CARD_DRAW;
		packet["player"] = (playerId == i);
		//if (playerId == i) { // "Hotfix"
			packet["cardId"] = card.id;
		//}

		sendNotify(i, packet);
	}
}

void GameSession::notifyUpdateCard(int playerId, int slotId, const Card& card) {
	for (int i = 0; i < 2; i++) {
		nlohmann::json statesPayload;
		statesPayload["hp"] = card.hp;
		statesPayload["ritual"] = card.rp;
		statesPayload["attack"] = card.attack;

		nlohmann::json packet;
		packet["id"] = PacketIdentifier::SERVER_UPDATE_CARD;
		packet["slot"] = slotId;
		packet["player"] = (playerId == i);
		packet["states"] = statesPayload;

		sendNotify(i, packet);
	}
}

void GameSession::sendCardDump(int playerId) {
	nlohmann::json packet;
	packet["id"] = PacketIdentifier::SERVER_CARD_INFODUMP;

	const std::map<int, Card>& cards = m_cardManager->getCards();

	nlohmann::json cardArray;

	for (auto it = cards.begin(); it != cards.end(); ++it){
		Card card = it->second;
		cardArray[card.id] = card.toJson();
	}

	packet["cardArray"] = cardArray;

	sendNotify(playerId, packet);
}

void GameSession::sendKeepalive(int playerId) {
	nlohmann::json packet;
	packet["id"] = PacketIdentifier::SERVER_KEEPALIVE;

	sendNotify(playerId, packet);
}
void GameSession::notifyPlayCard(int playerId, int targetPlayer, int fromSlot, int targetSlot, int clientMoveId) {
	for (int i = 0; i < 2; i++) {
		nlohmann::json packet;
		packet["id"] = PacketIdentifier::SERVER_PLAY_CARD;
		packet["sourcePlayer"] = (playerId == i);
		packet["targetPlayer"] = (targetPlayer == i);
		packet["sourceSlot"] = fromSlot;
		packet["targetSlot"] = targetSlot;
		if (playerId == i) {
			packet["moveId"] = clientMoveId;
		}

		sendNotify(i, packet);
	}
}

void GameSession::notifyDiscardCard(int playerId, int slot)
{
	for(int i=0; i<2; i++)
	{
		nlohmann::json packet;
		packet["id"] = PacketIdentifier::SERVER_DISCARD_CARD;
		packet["player"] = (playerId == i);
		packet["slot"] = slot;

		sendNotify(i, packet);
	}
}

void GameSession::notifyWinner(int playerId) {
	for (int i = 0; i<2; i++) {
		nlohmann::json packet;
		packet["id"] = PacketIdentifier::SERVER_WINNER_ID;
		packet["player"] = (playerId == i);

		sendNotify(i, packet);
	}
}

void GameSession::sendNotify(int playerId, const nlohmann::json& packet) {
	Player* player = players[playerId];
	if (player != nullptr) {
		player->networkSession->send(packet);
	}
}

void GameSession::notifyErrorMessage(int playerId, std::string error) {
	nlohmann::json packet;
	packet["id"] = PacketIdentifier::SERVER_ERROR_MESSAGE;
	packet["error"] = error;

	sendNotify(playerId, packet);
}

void GameSession::notifyErrorMessage(int playerId, std::string error, int clientMoveId) {
	nlohmann::json packet;
	packet["id"] = PacketIdentifier::SERVER_ERROR_MESSAGE;
	packet["error"] = error;
	packet["moveId"] = clientMoveId;

	sendNotify(playerId, packet);
}

session::session(tcp::socket socket, GameSession* p_gameSession) :
	socket_(std::move(socket)),
	timer_(socket_.get_io_service()),
	strand_(socket_.get_io_service()),
	m_gameSession(p_gameSession),
	m_player(this),
	m_messageQueue(socket_.get_io_service()),
	m_shuttingDown(false) {

}

// Initializes the session.
// Spawns two coroutines:
// * one to handle the requests and
// * one for waiting for timer related events
// The timer expiration does not interrupt the socket operations directly.
// Instead the socket is closed, causing the pending operations to awake and error out.
void session::go() {
	// The two coroutines both share the reference to the same session
	// This call is required to enable that?
	auto self(shared_from_this());

	bool canJoin = m_gameSession->joinGame(&m_player);

	if (canJoin == false) {
		std::cout << "Client denied. Game session full!" << std::endl;
		socket_.close();
		return;
	}

	std::cout << "Client: " << &socket_ << " Joined!" << std::endl;

	// Spawn the first coroutine
	// This performs the main functionality, reading data and echoing it back
	boost::asio::spawn(strand_,
		[this, self](boost::asio::yield_context yield) {
		try {
			// Allocate a small buffer for this connection
			char data[RECEIVE_BUFFER_SIZE];

			// Keep echoing data as long as data is provided in a timely manner
			for (;;) {
				// Client must send new data within two seconds or the connection is closed
				timer_.expires_from_now(std::chrono::seconds(ConfigManager::timeoutDuration));

				// Read 4 byte packet size from the socket to the buffer
				boost::asio::async_read(socket_, boost::asio::buffer(data), boost::asio::transfer_exactly(4), yield);

				// Interpret it as uint32
				uint32_t size = *((uint32_t*)&data);
				std::cout << "Size: " << size << std::endl;

				if (size > sizeof(data)) {
					std::cerr << "Too long, closing connection.\n";
//					timer_.cancel();
//					socket_.close();
					disconnect();
					return;

				}

				// Read the specified size
				boost::asio::async_read(socket_, boost::asio::buffer(data), boost::asio::transfer_exactly(size), yield);

				m_gameSession->handlepacket(&m_player, nlohmann::json::parse(std::string(data,size)));
			}
		}
		catch (std::exception& e) {
			// For example a socket error occurred
			// * write failed because socket was closed locally(or maybe remotely), or EOF
			//   * In these cases the error could be silently ignored, because this is somewhat expected.
			std::cerr << "Exception @read coroutine: " << e.what() << "\n";
			std::cerr << "ex type: " << typeid(e).name() << std::endl;
			std::cerr << "Socket id: " << &socket_ << std::endl;
			std::cerr << "Socket is open: " << socket_.is_open() << std::endl;

//			if (socket_.is_open() == true)
//				socket_.close();
//			timer_.cancel();
			disconnect();
		}
	});

		boost::asio::spawn(strand_,
			[this, self](boost::asio::yield_context yield) {
			try {
				while(1) {
					nlohmann::json packet = m_messageQueue.async_pop(yield);

					std::cout << "Received shit!" << packet << std::endl;

					if( packet.empty() ) {
						// TODO Exit this shit
						//std::cout << "Exiting this shit!" << std::endl;
//						socket_.close();
//						timer_.cancel();
						//disconnect();

						// Give the connection 2 seconds to finish data transfer
						// Arm a timer
						std::cout << "Empty detected, arming timer" << std::endl;
						boost::asio::steady_timer timer2(socket_.get_io_service());
						boost::system::error_code ignored_ec;
						timer2.expires_from_now(std::chrono::seconds(2));

						// Init shutdown
						std::cout << "Shutting down" << std::endl;
						boost::system::error_code ec;
						socket_.shutdown(tcp::socket::shutdown_both, ec);
						if (ec) {
							std::cerr << "Error: shutdown(): " << ec << std::endl;
						}

						std::cout << "Waiting" << std::endl;
						// Wait for some time
						timer2.async_wait(yield[ignored_ec]);
						std::cout << "Closing" << std::endl;
						// Free the socket resource and break the timer loop
						socket_.close();
						timer_.cancel();

						return;
					}
					else {
						//std::cout << "Sending shit back!";
						std::cout << "Sending to client: " << &socket_ << std::endl;
						std::string serialized = packet.dump();
						uint32_t size = serialized.size();
						boost::asio::async_write(socket_, boost::asio::buffer((char*)&size, 4), boost::asio::transfer_exactly(4), yield);
						boost::asio::async_write(socket_, boost::asio::buffer(serialized, size), boost::asio::transfer_exactly(size), yield);
					}
				}
			}
			catch (std::exception& e) {
				// For example a socket error occurred
				// * write failed because socket was closed locally(or maybe remotely), or EOF
				//   * In these cases the error could be silently ignored, because this is somewhat expected.
				std::cerr << "Exception: " << e.what() << "\n";
				std::cerr << "ex type: " << typeid(e).name() << std::endl;
//				socket_.close();
//				timer_.cancel();
				disconnect();
			}
		});

	boost::asio::spawn(strand_,
		[this, self](boost::asio::yield_context yield) {
		while (socket_.is_open()) {
			// Wait for something to happen to the timer
			boost::system::error_code ec;
			timer_.async_wait(yield[ec]);

			// Something happened. Not interested if the timer was cancelled.
			if (ec != boost::asio::error::operation_aborted) {

				if (!m_shuttingDown) {
					// Notify the client of the reason
					nlohmann::json packet;
					packet["id"] = "timeout";
					send(packet);
					disconnect();

					// Initiate graceful connection closure.
					// This ensures that all buffered data is sent to the client.
					// If this flushing is not required, close is enough.
					// The client will need to have some mechanism of detecting this, though.
					// Might want to replace shutdown this with SO_NOLINGER, non-zero timeout, and close?

					// For some reason this doesn't seem to work
					//boost::system::error_code ec;
					//socket_.shutdown(tcp::socket::shutdown_both, ec);
				}
				//else{
				//	socket_.close();
				//}
			}
		}
	});
}

void session::send(nlohmann::json packet) {
	if (m_shuttingDown && !packet.empty()) {
		std::cerr << "Tried to send data to a socket that is shutting down: " << packet << std::endl;
		return;
	}

	std::cout << "m_messageQueue.insert: " << packet << std::endl;
	m_messageQueue.insert(packet);
}

void session::disconnect() {
	if(socket_.is_open() && !m_shuttingDown) {
		m_shuttingDown = true;
		m_gameSession->playerDisconnected(&m_player);

		// Will eventually close the socket object
		send(nlohmann::json());
	}
}

void CardManager::loadCards(std::string filename) {
	// Load json from the filesystem
	std::ifstream ifs(filename);
	if (ifs.fail()) {
		throw myexception(strerror(errno));
	}

	// Parse the json object
	nlohmann::json cardsObject;
	cardsObject << ifs;

	// Loop the array and loads cards to the map
	int cumulativeChance = 0;
	for (auto cardObject : cardsObject) {
		Card card;
		card.readFromJson(cardObject);

		std::cout << card.name << ": " << card.flavorText << std::endl;
		cards[card.id] = card;

		cardIdsAndChances.push_back(std::pair<int, int>(card.id, cumulativeChance));
		cumulativeChance += card.chance;
	}

	std::cout << "Loaded " << cards.size() << " cards\n";

	// Also initialize the random generator
	uni = std::uniform_int_distribution<int>(0, cumulativeChance);
}

void Card::readFromJson(const nlohmann::json& cardObject) {
	this->id = cardObject["id"];
	this->name = cardObject["name"].get<std::string>();
	this->flavorText = cardObject["flavorText"].get<std::string>();
	this->cardTheme = cardObject["cardTheme"].get<std::string>();
	this->assetId = cardObject["assetId"];
	this->rarity = cardObject["rarity"];
	this->targetingFlags = std::bitset<sizeof(TargetingType)>((int)(cardObject["targetingFlags"]));
	this->chance = cardObject["chance"];
	this->type = cardObject["type"];

	if (this->type == (int)(CardType::Minion)) {
		this->attack = cardObject["attack"];
		this->hpMax = cardObject["hpMax"];
		this->hp = this->hpMax;
		this->rp = cardObject["ritual"];
	}
	else {
		this->attack = 0;
		this->hpMax = 0;
		this->hp = 0;
		this->rp = 0;
	}
	
	if (cardObject.count("script")) {
		this->script = cardObject["script"].get<std::string>();
	}
	else {
		this->script = std::string();
	}

	this->manaCost = cardObject["manaCost"];

	this->actionPoints = 0;
}

nlohmann::json Card::toJson() {
	nlohmann::json cardObject;

	cardObject["id"] = this->id;
	cardObject["name"] = this->name;
	cardObject["flavorText"] = this->flavorText;
	cardObject["cardTheme"] = this->cardTheme;
	cardObject["assetId"] = this->assetId;
	cardObject["rarity"] = this->rarity;
	//cardObject["targetingFlags"] = ???
	cardObject["attack"] = this->attack;
	cardObject["hp"] = this->hp;
	cardObject["hpMax"] = this->hpMax;
	cardObject["ritual"] = this->rp;
	cardObject["type"] = this->type;
	cardObject["manaCost"] = this->manaCost;

	return std::move(cardObject);
}

Card CardManager::getRandom() {
	int chanceNumber = uni(rng); // Get a random number

	// A safe default
	int selectedId = cardIdsAndChances[0].first;

	// Perform a slow linear lookup on chances
	for (auto& pair : cardIdsAndChances) {
		if (pair.second <= chanceNumber) {
			selectedId = pair.first;
		}
		else {
			break;
		}
	}

	return cards[selectedId];
}

Card CardManager::get(int id) {
	return cards[id];
}

int wrapped_main(int argc, char *argv[]) {
	printf("Server starting.\n");

	/*char *dirsep = strrchr(argv[0], '/');
	if (dirsep != NULL) *dirsep = 0;
	boost::filesystem::current_path(argv[0]);*/

	try {
		ConfigManager::loadConfig("../config.json");
		std::cout << "Config loaded\n";
	}
	catch (std::exception& e) {
		std::cerr << "Could not load config" << e.what() << "\n";
		std::cerr << "Using default values\n";
	}

	CardManager cardManager;
	cardManager.loadCards("../cards.json");	// TODO make variable or something?

	if (argc != 2) {
		std::cerr << "Usage: CardgameServer <port>\n";
		throw myexception("Usage: CardgameServer <port>");
	}

	// Parse the bind port number
	//int port = 1234;
	int port = std::atoi(argv[1]);

	printf("Listening at port %d\n", port);

	// Initialize the manager
	boost::asio::io_service io_service;

	GameSession theSession(&cardManager);

	// Spawn the TCP acceptor
	boost::asio::spawn(io_service, [&](boost::asio::yield_context yield) {
		// Create a new listening TCP socket that binds to the port given
		tcp::acceptor acceptor(io_service, tcp::endpoint(tcp::v4(), port));


		for (;;) { // Loop forever
			boost::system::error_code ec;
			tcp::socket socket(io_service); // Allocate a client socket
			acceptor.async_accept(socket, yield[ec]); // Try to accept a connection
			if (!ec) { // If no error
				// Create a new session instance and pass the client socket to it
				// Call the go-function, which creates new coroutines
				std::make_shared<session>(std::move(socket), &theSession)->go();
			}
			else {
				std::cout << "Error accepting: " << ec << "\n";
			}
		}
	});

	// Start executing coroutines
	io_service.run();

}


int main(int argc, char *argv[]) {
	try {
		return wrapped_main(argc, argv);
	}
	catch (std::exception& e) {
		std::cerr << "Exception in main: " << e.what() << "\n";
		return 1;
	}
}
