#include <boost/asio/io_service.hpp>
#include <boost/asio/ip/tcp.hpp>
#include <boost/asio/spawn.hpp>
#include <boost/asio/steady_timer.hpp>
#include <boost/asio/write.hpp>
#include <boost/asio/read.hpp>

#include <boost/interprocess/ipc/message_queue.hpp>

#include <boost/filesystem.hpp>

#include "json/json.hpp"

#include <random>
#include <memory>
#include <fstream>
#include <exception>
#include <bitset>

#include "queue.hpp"

using boost::asio::ip::tcp;

enum class CardRarity : int {
	COMMON = 0,
	RARE = 1,
	EPIC = 2,
	LEGENDARY = 3
};

enum class TargetingType : int {
	EnemyMinion = 1,
	EnemyPlayer = 2,
	OwnMinion = 4,
	OwnPlayer = 8
};

enum class CardType : int {
	Minion = 0,
	Spell = 1
};

class Card{
public:
	int id;

	std::string name;
	std::string flavorText;
	std::string cardTheme;
	std::string script;
	int type;
	int assetId;
	int rarity;
	std::bitset<sizeof(TargetingType)> targetingFlags;

	// std::vector<Enchanments> enchantments;

	int attack;
	int hp;
	int hpMax;
	int rp;
	int actionPoints;
	int manaCost;

	int chance;

	void readFromJson(const nlohmann::json& cardObject);
	nlohmann::json toJson();
};


//class Board{
//	Card backline[5];
//	Card frontline[4];
//};

class session; // fwd decl

class Player{
	public:
		int ritualPoints;
		int mana;

		int serverSlot; // 0 or 1

//		Board board;
//		std::vector<Card> deck;
//		std::vector<Card> field;	// Cards that are on the table (0<=id<=7) or in the hand (8<id)

//		std::shared_ptr<session> networkSession;
	public:
		std::vector<Card> deck;
		std::vector<Card> hand;
		std::vector<Card> field;	// Cards that are on the table (0<=id<=7) or in the hand (8<id)

		session* networkSession;
		Player(session* p_session);
};

enum class EGameState
{
	WAIT_FOR_PLAYERS,
	INITIALIZE_GAME,
	PLAYER_TURN_START,
	PLAYER_ACTION,
	PLAYER_TURN_END,
	CHANGE_TURN,
	GAME_END,
	UNITIALIZE_GAME,
};

class CardManager;
class GameSession
{
	private:
		Player *players[2];
		int currentPlayer;

		EGameState m_currentState;

		CardManager* m_cardManager;

		// Sends a packet to both players
		void notifyUpdatePlayer(int playerId, Player* player);
		void notifyJoined(int playerId);
		void notifyTurnStart(int playerId);
		void notifyTurnEnd(int playerId);
		void notifyCardDraw(int playerId, const Card& card);
		void notifyWinner(int playerId);
		void notifyUpdateCard(int playerId, int slotId, const Card& card);
		void notifyPlayCard(int srcPlayer, int trgPlayer, int srcSlot, int trgSlot, int clientMoveId);
		void notifyDiscardCard(int Player, int Slot);

		void notifyErrorMessage(int playerId, std::string error);
		void notifyErrorMessage(int playerId, std::string error, int clientMoveId);

		// Sends a packet to just a single player
		void sendCardDump(int playerId);
		void sendKeepalive(int playerId);

		// Sends a message to the given player
		void sendNotify(int playerId, const nlohmann::json& packet);

//		boost::asio::steady_timer turnTimer;
	public:
		GameSession(CardManager* p_cardManager);

		void playerDisconnected(Player* p_player);

		void handlepacket(Player* p_player, nlohmann::json json_packet);

		bool joinGame(Player* player);
		bool isGameFull() { return (players[0] != nullptr && players[1] != nullptr); };
		void startGame();

		void playerStartTurn();
		void playerActions(const std::string& action, nlohmann::json json_string);
		void playerEndTurn();

		void changeTurn();

		void gameEnd();
		void cleanup();
};

//
// Joku tulevaisuuden härdeli
//

class game_manager{
	// In the future support multiple sessions
	GameSession theSession;
};



class session : public std::enable_shared_from_this<session> {
private:
	tcp::socket socket_;

	boost::asio::steady_timer timer_;
	boost::asio::io_service::strand strand_;

	GameSession *m_gameSession;		// The game session this client is part of
	Player m_player;				// The player object of this session;

	simple_queue m_messageQueue;
	bool m_shuttingDown;

public:
	explicit session(tcp::socket socket, GameSession* p_gameSession);

	// Initializes the session.
	// Spawns two coroutines:
	// * one to handle the requests and
	// * one for waiting for timer related events
	// The timer expiration does not interrupt the socket operations directly.
	// Instead the socket is closed, causing the pending operations to awake and error out.
	void go();
	void send(nlohmann::json buffer);

	void disconnect();
};


class CardManager {
	private:
		std::map<int, Card> cards;
		std::vector<std::pair<int, int>> cardIdsAndChances;

		// random-number engine used (Mersenne-Twister in this case)
		std::mt19937 rng;
		std::uniform_int_distribution<int> uni;

	public:
		CardManager() : rng(std::random_device()()){}

		void loadCards(std::string filename);
		Card getRandom();
		Card get(int id);

		const std::map<int, Card>& getCards() { return cards; }
};


class myexception : public std::exception{
	private:
		std::string text;
	public:
		myexception(std::string txt) : text(txt) {}

		virtual const char* what() const throw(){
			return "My exception happened";
		}
};
